#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `transit_chem` package."""


from click.testing import CliRunner

import transit_chem
from transit_chem import cli


def test_content():
    transit_chem.HarmonicOscillator


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.main)
    assert result.exit_code == 0
    help_result = runner.invoke(cli.main, ["--help"])
    assert help_result.exit_code == 0



import itertools
import logging
from abc import ABC

import numpy as np
from scipy import linalg

from transittime.log_helper import TimeLevel, log_block

mod_log = logging.getLogger(__name__)


class TransitTime(ABC):
    """The transit time Class. This is the base class from which OneD and
    ThreeD inherit. The time-dependence is all in this class. The details
    of spatially dependent quantities, like potentials and integrators are
    defined in the child classes.

    Metaclass=ABCMeta defines this class as an Abstract base class. This means it can't be directly
    instantiated and only one of the subclasses can be instantiated. This is to protect you from
    yourself.

    """

    @log_block("Transittime Init", mod_log, logging.DEBUG)
    def __init__(self, nregions):
        """Initializer for Transit time object.

        :param nregions: Number of spatially separated regions to consider.
        :type nregions: int
        """
        self.Nregions = nregions
        self.times = np.zeros(0)
        self.OccProb = np.zeros((0, 0))
        self.ChangeMat = np.zeros((0, 0, 0))
        self.S_partial = np.zeros((0, 0, 0))
        self.S = np.zeros((0, 0))
        self.T = np.zeros((0, 0))
        self.V = np.zeros((0, 0))
        self.H = np.zeros((0, 0))
        self.eigvals = np.zeros((0, 0))
        self.eigvecs = np.zeros((0, 0))
        self.eigenbasis = np.zeros((0, 0))
        self.P_not = np.zeros(0)
        self.Po = np.zeros(0)
        self.basis_vec = []
        self.Nbasis = 0
        self.state_mo_coefficients = None

    @log_block(
        "Calculating both time independent and time-dependent quantities.",
        mod_log,
        logging.INFO,
        TimeLevel.ALWAYS,
    )
    def calc_all(self, initialstate):
        """Calculate both time independent and time-dependent quantities.

        :param initialstate: The initial state of the system.
        :type initialstate: (3D) function of cartesian coordinates f(x, y, z).
                            (1D) Array containing the values of initial state on a grid.
        :param times: NumPy array of time steps to be used.
        :type times: ndarray
        """
        self.calc_time_independent()
        self.set_initial_state(initialstate)
        self.calc_time_dependent()

    @log_block(
        "Calculate all the time independent quantities",
        mod_log,
        logging.INFO,
        TimeLevel.OFTEN,
    )
    def calc_time_independent(self):
        """Calculate all the time independent quantities.

        Calculates the eigenbasis as well as partial overlap matrix.
        """
        self.diagonalize_H()
        self.get_eigen_basis()
        self.calc_partial_overlap()

    @log_block(
        "Calculate all the time dependent quantities",
        mod_log,
        logging.INFO,
        TimeLevel.OFTEN,
    )
    def calc_time_dependent(
        self, guess_t_max=500, p_not_threshold=0.05, stepsize=1.0, maxsteps=10 ** 4
    ):
        """Calculate the time dependent quantities. Must be after time-independent is done.

        Calculates occupancy probabilities, the change matrix and pnot.

        """

        # Recalculate until P_not < threshold or  n_steps = t_max // stepsize >
        p_not = 1.0
        t_max = guess_t_max
        n_steps = 1
        while (p_not >= p_not_threshold) and (n_steps < maxsteps):
            n_steps = round(t_max / stepsize)
            times = np.linspace(0, t_max, n_steps)
            self.times = times
            self.calc_occupancy_probability()
            self.get_change_mat()
            self.get_p_not(3)
            p_not = self.P_not[-2]
            t_max *= 2

    def set_initial_state(self, initialstate):
        raise NotImplementedError

    def calc_partial_overlap(self):
        raise NotImplementedError

    def normalize(self, vec):
        raise NotImplementedError

    def get_S(self):
        raise NotImplementedError

    def get_H(self):
        raise NotImplementedError

    def vec_to_ao(self, vec):
        raise NotImplementedError

    def vec_to_mo(self, vec):
        raise NotImplementedError

    def propagate_state(self, state):
        raise NotImplementedError

    def get_eigen_basis(self):
        raise NotImplementedError

    @log_block(
        "Diagonalize the Hamiltonian",
        mod_log,
        level=logging.INFO,
        time_level=TimeLevel.RARELY,
    )
    def diagonalize_H(self):
        """Calculate H, S, V, T and diagonalize. Form eigenbasis.

        """
        # Get matrices
        self.get_S()
        self.get_H()

        # Diagonalize and sort
        self.eigvals, self.eigvecs = linalg.eigh(self.H, self.S)
        idx = np.argsort(self.eigvals)
        self.eigvals = self.eigvals[idx]
        self.eigvecs = self.eigvecs[:, idx]

        #  Normalize eigenvectors
        norm = np.linalg.norm(self.eigvecs, axis=0)
        self.eigvecs /= norm[None, :]

    @log_block("ao to mo transformation", mod_log, level=logging.DEBUG)
    def ao_to_mo(self, ao_vec):
        """Given a vector of coefficients in AO basis, return coefficients in MO basis."""
        mo = np.dot(self.eigvecs.T, ao_vec)
        mo /= np.linalg.norm(mo)
        return mo

    @log_block("Mo to ao transformation", mod_log, level=logging.DEBUG)
    def mo_to_ao(self, mo_vec):
        """Given a vector of coefficients in MO basis, return coefficients in AO basis."""
        ao = np.dot(self.eigvecs, mo_vec)
        ao /= np.linalg.norm(ao)
        return ao

    @log_block("Matrix ao to mo transformation", mod_log, level=logging.DEBUG)
    def matrix_ao_to_mo(self, matrix):
        inv = np.linalg.inv
        return inv(self.eigvecs) @ inv(self.S) @ matrix @ self.eigvecs

    @log_block("Determine the occupancy probability", mod_log, level=logging.INFO)
    def calc_occupancy_probability(self):
        """Calculate the time-dependent occupancy probabilities"""
        c_conjugate = []
        c = []
        for i in range(self.Nbasis):
            c.append(
                self.state_mo_coefficients[i]
                * np.exp(-1j * self.eigvals[i] * self.times)
            )
            c_conjugate.append(
                self.state_mo_coefficients[i]
                * np.exp(1j * self.eigvals[i] * self.times)
            )

        occ_prob = np.zeros((self.Nregions, len(self.times)), dtype=np.complex128)

        for i in range(self.Nregions):
            for n in range(self.Nbasis):
                for m in range(self.Nbasis):
                    occ_prob[i] += c_conjugate[n] * c[m] * self.S_partial[i, n, m]
        self.OccProb = np.abs(np.real(occ_prob))
        self.OccProb /= np.sum(self.OccProb, axis=0)  # Normalize
        return self.OccProb

    @log_block("Determine the change/transition matrix", mod_log, level=logging.INFO)
    def get_change_mat(self):
        """calculate the change matrix over time, currently supports 3 regions"""
        A = np.zeros(
            (len(self.times) - 1, self.Nregions, self.Nregions)
        )  # change matrix
        dOccProb = [np.diff(i) for i in self.OccProb]  # change for each region
        sign_of_change = [np.sign(i) for i in dOccProb]
        two_states_increasing = (
            sign_of_change[0] * sign_of_change[1] * sign_of_change[2] < 0
        )

        for t in range(len(self.times) - 1):
            if two_states_increasing[t]:
                # If the state is decreasing in probability, the diagonal is
                # determined
                for j in range(3):
                    if sign_of_change[j][t] < 0:
                        A[t, j, j] = self.OccProb[j, t + 1] / self.OccProb[j, t]
                        for k in range(3):
                            if k != j:
                                A[t, k, j] = dOccProb[k][t] / self.OccProb[j, t]

                    # If the state is increasing, diagonal is 1. Off diagonal
                    # columns = 0
                    elif sign_of_change[j][t] > 0:
                        A[t, j, j] = 1
                        for k in range(3):
                            if k != j:
                                A[t, k, j] = 0

            elif not two_states_increasing[t]:
                for j in range(3):
                    # If the state is decreasing in probability, the diagonal
                    # is determined
                    if sign_of_change[j][t] < 0:
                        A[t, j, j] = self.OccProb[j, t + 1] / self.OccProb[j, t]
                        for k in range(3):
                            if k != j:
                                A[t, j, k] = 0

                # If the state is increasing, diagonal is 1. Off diagonal
                # column element is 1 - diagonal
                for j in range(3):
                    if sign_of_change[j][t] > 0:
                        A[t, j, j] = 1
                        for k in range(3):
                            if k != j:
                                A[t, j, k] = 1 - A[t, k, k]

            else:  # just in case weird equivalence happens, repeat last time step
                A[t, :, :] = A[t - 1, :, :]
        self.ChangeMat = A
        return A

    @log_block(
        "Determine p_not - this gets called multiple times until it finds desired threshold",
        mod_log,
        level=logging.INFO,
        time_level=TimeLevel.OFTEN,
    )
    def get_p_not(self, acceptor):
        """Determine P_not, the probability than acceptor state has never been occupied.

        :param acceptor: index of the state to consider.
        :type acceptor: int
        TODO: GET ACCEPTOR/DONOR
        """
        self.Po = self.OccProb[:, 0]

        #  Make mask to remove the acceptor rows and columns
        mask = np.full_like(self.ChangeMat, True, dtype=bool)
        mask[:, acceptor - 1, :] = False
        mask[:, :, acceptor - 1] = False
        newshape = self.ChangeMat.shape
        newshape = newshape[0], newshape[1] - 1, newshape[2] - 1
        A_tilde = self.ChangeMat[mask].reshape(newshape)

        P_o_tilde = self.Po[:2]
        #  Taking [a, b, c, d, ...] -> [(b @ a), (c @ b @ a), (d @ c @ b @ a), ...] and making into an array.
        a_prod = np.asarray(list(itertools.accumulate(A_tilde, lambda a, b: b @ a)))
        #  Broadcasting rules means that each 2x2 submatrix of a_prod matrix multiplies P_o_tilde
        self.P_not = np.sum(a_prod @ P_o_tilde, axis=1)



import copy
import glob
import sys

import numpy as np

from . import atom


class Molecule(object):
    """Molecule class. Contains list of atoms as well as collective properties.

    Contains functions for rotating, aligning and scaling. Iterating over the molecule gives each atom.
    """

    def __init__(self, xyz_file, basis_type="minimalpz"):
        """Constructor for Molecule from an xyzfile.

        The xyzfile can be either of these forms:

        C 1.0  1.0  1.0
        H 2.0 2.0 2.0
        ...

        or

        6 1.0 1.0 1.0
        1 1.0 1.0 1.0
        ...

        :param xyz_file: Name of xyzfile containing atoms and coordinates
        :type xyz_file: str
        :param basis_type: Descriptor for basis function. Currently supports
        only one type across all atom types.
        :type basis_type: str
        """
        self.xyzfile = xyz_file
        self.comment = ""
        self._atoms_saved, self.comment = self.xyz_to_atoms(self.xyzfile, basis_type)
        self.atoms = copy.deepcopy(self._atoms_saved)

    @property
    def Natoms(self):
        return len(self.atoms)

    @property
    def coords(self):
        return np.asarray([atom.coords for atom in self.atoms])

    def trim_hydrogens(self):
        self.atoms = [atom for atom in self._atoms_saved if not atom.symbol == "H"]

    def show_hydrogens(self):
        self.atoms = self._atoms_saved

    def __iter__(self):
        self._gen = iter(self.atoms)
        return self

    def __next__(self):
        return next(self._gen)

    @property
    def R(self):
        """Calculate the distances for each atom-atom pair."""
        r = np.zeros((self.Natoms, self.Natoms))
        for i in range(self.Natoms):
            for j in range(self.Natoms):
                diffs = np.asarray(self.atoms[i].coords) - np.asarray(
                    self.atoms[j].coords
                )
                r[i, j] = np.sqrt(np.sum(diffs ** 2))
        return r

    @property
    def center_of_mass(self):
        """Determine the center of mass of the molecule."""
        total_mass = 0.0
        mx = 0.0
        my = 0.0
        mz = 0.0
        for a in self.atoms:
            total_mass += a.Z
            mx += a.Z * a.coords[0]
            my += a.Z * a.coords[1]
            mz += a.Z * a.coords[2]
        return mx / total_mass, my / total_mass, mz / total_mass

    def translate(self, x, y, z):
        for a in self.atoms:
            a.translate(x, y, z)

    @staticmethod
    def xyz_to_atoms(xyz_file, basis_type):
        """Convert from xyzfile and basistype into list of atoms.

        :param xyz_file: Name of xyzfile containing atoms and coordinates
        :type xyz_file: str
        :param basis_type: Descriptor for basis function. Currently supports
        only one type across all atom types.
        :type basis_type: str
        """
        with open(xyz_file) as f:
            text = f.readlines()
            comment = [line.rstrip() for line in text[:2]]
            text = text[2:]  # First two are number and comment lines
            atomlist = [line.rstrip().split() for line in text]
            atomtypes, coords = zip(*[(line[0], line[1:]) for line in atomlist])
            coords = [[float(c) for c in line] for line in coords]

        atoms = []
        for i, atomtype in enumerate(atomtypes):
            atoms.append(atom.Atom(atomtype, coords[i], basis_type))
        return atoms, comment

    def align_to_x(self, atom1, atom2, xy_anchor=None):
        """Align the line connecting atom1 and atom2 to the x axis and set the midpoint to the origin.

        after aligning atom1 and atom2 to x axis, rotate the molecule about the x axis such that
        the xy_anchor atom lies in the xy plane.
        """
        midpoint = [
            0.5 * (coord2 + coord1)
            for coord1, coord2 in zip(atom1.coords, atom2.coords)
        ]
        self.translate(-midpoint[0], -midpoint[1], -midpoint[2])

        #  Rotating atom1 to be in the xz plane
        x, y, z = atom1.coords
        theta = -np.arctan(y / x)
        Rz = np.asarray(
            [
                [np.cos(theta), -np.sin(theta), 0],
                [np.sin(theta), np.cos(theta), 0],
                [0, 0, 1],
            ]
        )
        for a in self.atoms:
            a.rotate(Rz)

        # Rotate to put atom1 in on the +x axis from within xz plane
        x, y, z = atom1.coords
        theta = -np.arctan(z / x)
        Ry = np.asarray(
            [
                [np.cos(theta), 0, -np.sin(theta)],
                [0, 1, 0],
                [np.sin(theta), 0, np.cos(theta)],
            ]
        )
        for a in self.atoms:
            a.rotate(Ry)

        # Rotate to put xy anchor on the xy plane
        if xy_anchor is not None:
            x, y, z = xy_anchor.coords
            theta = -np.arctan(z / y)
            Rx = np.asarray(
                [
                    [1, 0, 0],
                    [0, np.cos(theta), -np.sin(theta)],
                    [0, np.sin(theta), np.cos(theta)],
                ]
            )
            for a in self.atoms:
                a.rotate(Rx)

    def scale(self, factor):
        for a in self.atoms:
            a.scale(factor)

    def flip_x(self):
        for a in self.atoms:
            a.flip_x()

    def sort(self, atomic_key):
        self.atoms = sorted(self.atoms, key=atomic_key)

    def __str__(self):
        my_list = self.comment + [str(atom) for atom in self.atoms]
        s = "\n".join(my_list)
        return s


def main(argv):
    for f in glob.glob("../analysis/xyz_files/*.xyz"):
        M = Molecule(f)
        M.trim_hydrogens()
        nitrogens = [atom for atom in M.atoms if atom.Z == 7]
        if nitrogens:
            M.align_to_x(*nitrogens)

        #  Align the nitro (acceptor) group to be +x
        oxygens = [atom for atom in M.atoms if atom.Z == 8]
        if oxygens:
            if all(oxygen.coords[0] < 0 for oxygen in oxygens):
                M.flip_x()
        M.sort(atomic_key=lambda atom: atom.coords[0])
        print(M)
        print(M.coords)


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import io
import os
from pathlib import Path

from setuptools import find_packages, setup

# Package meta-data.
NAME = "transit_chem"
DESCRIPTION = "A Transit chem"
URL = "https://github.com/ecurtin2/transit-chem"
EMAIL = "evanmcurtin@gmail.com"
AUTHOR = "Evan Curtin"
REQUIRES_PYTHON = ">=3.4.0"
VERSION = "0.5.0"

# What packages are required for this module to be executed?
this_dir = Path(__file__).parent
REQUIRED = (this_dir / "requirements.txt").read_text().split()

# What packages are optional?
EXTRAS_DOC = (this_dir / "docs/requirements_doc.txt").read_text().split()
EXTRAS_DEV = (this_dir / "requirements_dev.txt").read_text().split()

EXTRAS = {"dev": EXTRAS_DEV, "doc": EXTRAS_DOC, "all": EXTRAS_DEV + EXTRAS_DOC}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


# Where the magic happens:
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],
    # entry_points={
    #     'console_scripts': ['mycli=mymodule:cli'],
    # },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)



import numpy as np

import transittime


def main(argv=None):

    x = np.linspace(-10, 10, 500)
    one = transittime.OneD(coordinates=x, nregions=3)
    # t = np.linspace(0, 1500, 1500)
    initial_state = one.basis_funcs[0](one.coords, 0, 0.0)
    one.calc_all(initialstate=initial_state)

    # fname = '../analysis/xyz_files/2C.xyz'
    # mol = transittime.Molecule(fname)
    # mol.trim_hydrogens()
    # mol.scale(transittime.units.conversion_factors['ANGSTROM_TO_BOHR'])

    #  Aligns the line joining the two nitrogens with the x axis.
    # nitrogens = [atom for atom in mol if atom.symbol == 'N']
    # if nitrogens:
    #    mol.align_to_x(*nitrogens)
    # else:
    #    raise AttributeError('Did not find nitrogens!')

    #  Align the nitro (acceptor) group to be +x
    # oxygens = [atom for atom in mol if atom.symbol == 'O']
    # if oxygens:
    #    if all(oxygen.coords[0] < 0 for oxygen in oxygens):
    #        mol.flip_x()

    #  Sort into ascending order of x coordinates for convenience
    # mol.sort(atomic_key=lambda atom: atom.x)

    #  This works since the molecule is aligned/shifted as expected.
    # donor_n = min(nitrogens, key=lambda atom: atom.x)
    # acceptor_n = max(nitrogens, key=lambda atom: atom.x)
    # donor_index = mol.atoms.index(donor_n)
    # acceptor_index = mol.atoms.index(acceptor_n)

    #  Since they are aligned on the x axis already and atoms are sorted by x value, the adjacent atoms in
    #  the list will be the relevant carbons for the boundary

    # donor_c = mol.atoms[donor_index + 1]
    # acceptor_c = mol.atoms[acceptor_index - 1]

    # donor_bridge_bound = 0.5 * (donor_n.x + donor_c.x)
    # acceptor_bridge_bound = 0.5 * (acceptor_n.x + acceptor_c.x)
    # bounds = (donor_bridge_bound, acceptor_bridge_bound)

    # ThreeDCalc = transittime.threeD.ThreeD(nregions=3, boundaries=bounds, molecule=mol, id='testing')
    # ThreeDCalc.calc_time_independent()

    #  Initial state is Pz orbital on donor nitrogen. This was most convenient way to do it.
    # initialstate = transittime.basis.PzBasis(*donor_n.coords, donor_n.alpha)
    # ThreeDCalc.set_initial_state(initialstate)  # does overlap integrals, can be slow
    # ThreeDCalc.calc_time_dependent(maxsteps=10**6)


if __name__ == "__main__":
    import sys

    main(sys.argv)
